# -*- coding: utf-8 -*-
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""Advanced example technology module"""

import yaml
import os

# Technology parameters
# should be tech_param_path = "laygo2_tech/laygo2_tech.yaml" for actual use
# tech_param_path = "laygo2/examples/laygo2_tech/laygo2_tech.yaml"
tech_param_path = "laygo2_tech_quick_start/laygo2_tech.yaml"
"""
path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
tech_param_path = dir_path+"/laygo2_tech.yaml"
"""

with open(tech_param_path, "r") as stream:
    try:
        tech_params = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)

from .laygo2_tech_templates import load_templates
from .laygo2_tech_grids import load_grids



# -*- coding: utf-8 -*-
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""Template library for the advanced example technology. (advtech)"""

import numpy as np

import laygo2.object.template
import laygo2.object.physical
import laygo2.object.database

# Should be `import laygo2_tech as tech` for actual use.
# import laygo2_tech as tech
import laygo2_tech_quick_start as tech

tech_params = tech.tech_params
# templates = tech_params["templates"]
grids = tech_params["grids"]

# Template functions for primitive devices
def _mos_update_params(params):
    """Make a complete parameter table for mos"""
    us = np.array([30, 100])  # unit transistor size
    ush = np.array([15, 100])  # half-unit transistor size
    if "nf" not in params:  # number of fingers
        params["nf"] = 1
    if "nfdmyl" not in params:  # number of left-dummy fingers
        params["nfdmyl"] = 0
    if "nfdmyr" not in params:  # number of right-dummy fingers
        params["nfdmyr"] = 0
    if "trackswap" not in params:  # source-drain track swap
        params["trackswap"] = False
    if "tie" not in params:  # tie to power rail
        params["tie"] = None
    if "bndl" not in params:  # create local left boundary
        params["bndl"] = True
    if "bndr" not in params:  # create local right boundary
        params["bndr"] = True
    if "gbndl" not in params:  # create global left boundary
        params["gbndl"] = False
    if "gbndr" not in params:  # create global right boundary
        params["gbndr"] = False
    if "unit_size_core" not in params:  # core unit size
        params["unit_size_core"] = us
    if "unit_size_dmy" not in params:  # dummy size
        params["unit_size_dmy"] = us
    if "unit_size_bndl" not in params:  # left boundary unit size
        params["unit_size_bndl"] = ush
    if "unit_size_bndr" not in params:  # right boundary unit size
        params["unit_size_bndr"] = ush
    if "unit_size_gbndl" not in params:  # left boundary unit size
        params["unit_size_gbndl"] = ush
    if "unit_size_gbndr" not in params:  # right boundary unit size
        params["unit_size_gbndr"] = ush
    return params


def mos_bbox_func(params):
    """Computes x and y coordinate values from params."""
    params = _mos_update_params(params)
    unit_size = params["unit_size_core"]
    xy0 = [0, 0]
    xy1 = [unit_size[0] * params["nf"], unit_size[1]]
    xy = np.array([xy0, xy1])
    # Increment bbox based on boundary / dummy parameters.
    if params["gbndl"]:
        xy[1, 0] += params["unit_size_gbndl"][0]
    if params["bndl"]:
        xy[1, 0] += params["unit_size_bndl"][0]
    if params["nfdmyl"] > 0:
        xy[1, 0] += params["unit_size_dmy"][0] * round(params["nfdmyl"])
    if params["nfdmyr"] > 0:
        xy[1, 0] += params["unit_size_dmy"][0] * round(params["nfdmyr"])
    if params["bndr"]:
        xy[1, 0] += params["unit_size_bndr"][0]
    if params["gbndr"]:
        xy[1, 0] += params["unit_size_gbndr"][0]
    return xy


def mos_pins_func(devtype, params):
    """Generate a pin dictionary from params."""
    # Compute parameters
    params = _mos_update_params(params)
    nf = params["nf"] if "nf" in params.keys() else 1
    nfdl = params["nfdmyl"]
    nfdr = params["nfdmyr"]
    blx = 15 if params["bndl"] == True else 0
    brx = 15 if params["bndr"] == True else 0
    gblx = 15 if params["gbndl"] == True else 0
    gbrx = 15 if params["gbndr"] == True else 0
    us = np.array([30, 100])  # unit transistor size

    # Create a pin dictionary
    pins = dict()
    # metal2 - gate
    pxy = np.array([[gblx + blx + 12, 85], [gblx + blx + us[0] * nf - 12, us[1] - 5]])
    pins["G"] = laygo2.object.Pin(xy=pxy, layer=["metal2", "drawing"], netname="G")
    # metal2 - sourcedrain
    srx = -1 * (nf % 2) * us[0]
    if params["trackswap"] == False:
        sy = 35
        dy = 55
    else:
        sy = 55
        dy = 35
    sdx = gblx + blx + nfdl * us[0]
    pxy = np.array([[sdx - 7, sy], [sdx + us[0] * nf + srx + 7, sy + 10]])
    pins["S"] = laygo2.object.Pin(xy=pxy, layer=["metal2", "drawing"], netname="S")
    drx = -1 * ((nf + 1) % 2) * us[0]
    pxy = np.array([[sdx + us[0] - 7, dy], [sdx + us[0] * nf + drx + 7, dy + 10]])
    pins["D"] = laygo2.object.Pin(xy=pxy, layer=["metal2", "drawing"], netname="D")
    # metal2 - rail
    pxy = np.array([[0, -5], [gblx + blx + us[0] * (nfdl + nf + nfdr) + brx + gbrx, 5]])
    pins["RAIL"] = laygo2.object.Pin(xy=pxy, layer=["metal2", "drawing"], netname="RAIL")
    return pins


def pmos_pins_func(params):
    return mos_pins_func(devtype="pmos", params=params)


def nmos_pins_func(params):
    return mos_pins_func(devtype="nmos", params=params)


def mos_generate_func(devtype, name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    """Generates an instance from the input parameters."""
    # Compute parameters
    params = _mos_update_params(params)
    nf = params["nf"] if "nf" in params.keys() else 1
    nfdl = params["nfdmyl"]
    nfdr = params["nfdmyr"]
    blx = 15 if params["bndl"] == True else 0
    brx = 15 if params["bndr"] == True else 0
    gblx = 15 if params["gbndl"] == True else 0
    gbrx = 15 if params["gbndr"] == True else 0
    us = np.array([30, 100])  # unit transistor size

    # Create the base mosfet structure.
    nelements = dict()
    # prboundary
    rxy = np.array([[0, 0], [gblx + blx + us[0] * (nfdl + nf + nfdr) + brx + gbrx, us[1]]])
    nelements["PRBND0"] = laygo2.object.Rect(xy=rxy, layer=["prBoundary", "drawing"], name="PRBND0")
    # well
    rxy = np.array([[-20, 0], [gblx + blx + us[0] * (nfdl + nf + nfdr) + brx + gbrx + 20, us[1]]])
    nelements["NW0"] = laygo2.object.Rect(xy=rxy, layer=["nwell", "drawing"], name="NW0")
    # implant
    rxy = np.array([[0, 5], [gblx + blx + us[0] * (nfdl + nf + nfdr) + brx + gbrx, us[1] - 5]])
    if devtype == "nmos":
        nelements["NIMPL0"] = laygo2.object.Rect(xy=rxy, layer=["nimplant", "drawing"], name="NIMPL0")
    else:
        nelements["PIMPL0"] = laygo2.object.Rect(xy=rxy, layer=["pimplant", "drawing"], name="PIMPL0")
    # diffusion
    rxy = np.array([[gblx + blx - 5, 25], [gblx + blx + us[0] * (nfdl + nf + nfdr) + 5, us[1] - 25]])
    nelements["DIFF0"] = laygo2.object.Rect(xy=rxy, layer=["diffusion", "drawing"], name="DIFF0")
    # poly
    for i in range(nf + nfdl + nfdr):
        rxy = np.array([[gblx + blx + us[0] * i + 12, 10], [gblx + blx + us[0] * (i + 1) - 12, us[1] - 10]])
        nelements["POLY" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["poly", "drawing"], name="POLY" + str(i))
    # metal1 - gate
    rxy = np.array([[gblx + blx + 12, 85], [gblx + blx + us[0] * (nfdl + nf + nfdr) - 12, us[1] - 5]])
    nelements["M1G0"] = laygo2.object.Rect(xy=rxy, layer=["metal1", "drawing"], name="M1G0")
    # metal1 - sourcedrain
    for i in range(nf + nfdl + nfdr + 1):
        rxy = np.array([[gblx + blx + us[0] * i - 7, 23], [gblx + blx + us[0] * i + 7, us[1] - 23]])
        nelements["M1SD" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["metal1", "drawing"], name="M1SD" + str(i))
    # via1 - gate
    gx = gblx + blx + nfdl * us[0]
    for i in range(nf):
        rxy = np.array([[gx + us[0] * i - 5 + 15, 85], [gx + us[0] * i + 5 + 15, us[1] - 5]])
        nelements["VG0" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["via1", "drawing"], name="VG0" + str(i))
    # via1 - sourcedrain
    if params["trackswap"] == False:
        sy = 35
        dy = 55
    else:
        sy = 55
        dy = 35
    sdx = gblx + blx + nfdl * us[0]
    for i in range(nf + 1):
        if i % 2 == 0:
            rxy = np.array([[sdx + us[0] * i - 5, sy], [sdx + us[0] * i + 5, sy + 10]])
            nelements["VS0" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["via1", "drawing"], name="VS0" + str(i))
        else:
            rxy = np.array([[sdx + us[0] * i - 5, dy], [sdx + us[0] * i + 5, dy + 10]])
            nelements["VD0" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["via1", "drawing"], name="VD0" + str(i))
    # metal1/via1 - tie
    sdx = gblx + blx + nfdl * us[0]
    for i in range(nf + 1):
        tie = False
        if (i % 2 == 0) and (params["tie"] is not None):
            if params["tie"] == "S":
                tie = True
        if (i % 2 == 1) and (params["tie"] is not None):
            if params["tie"] == "D":
                tie = True
        if tie:
            rxy = np.array([[sdx + us[0] * i - 7, -5], [sdx + us[0] * i + 7, 23]])
            nelements["M1TIE" + str(i)] = laygo2.object.Rect(
                xy=rxy, layer=["metal1", "drawing"], name="M1TIE" + str(i)
            )
            rxy = np.array([[sdx + us[0] * i - 5, -5], [sdx + us[0] * i + 5, 5]])
            nelements["V1TIE" + str(i)] = laygo2.object.Rect(xy=rxy, layer=["via1", "drawing"], name="V1TIE" + str(i))
    # metal2 - gate
    rxy = np.array([[gblx + blx + 12, 85], [gblx + blx + us[0] * nf - 12, us[1] - 5]])
    nelements["M2G0"] = laygo2.object.Rect(xy=rxy, layer=["metal2", "drawing"], name="M2G0")
    # metal2 - sourcedrain
    srx = -1 * (nf % 2) * us[0]
    rxy = np.array([[sdx - 7, sy], [sdx + us[0] * nf + srx + 7, sy + 10]])
    nelements["M2S0"] = laygo2.object.Rect(xy=rxy, layer=["metal2", "drawing"], name="M2S0")
    drx = -1 * ((nf + 1) % 2) * us[0]
    rxy = np.array([[sdx + us[0] - 7, dy], [sdx + us[0] * nf + drx + 7, dy + 10]])
    nelements["M2D0"] = laygo2.object.Rect(xy=rxy, layer=["metal2", "drawing"], name="M2D0")
    # metal2 - rail
    rxy = np.array([[0, -5], [gblx + blx + us[0] * (nfdl + nf + nfdr) + brx + gbrx, 5]])
    nelements["M2RAIL0"] = laygo2.object.Rect(xy=rxy, layer=["metal2", "drawing"], name="M2RAIL0")

    # Create pins
    pins = mos_pins_func(devtype=devtype, params=params)
    nelements.update(pins)  # Add physical pin structures to the virtual object.

    # inst_unit_size = unit_size * np.array([nf, 1])
    # Unit size
    inst_xy = mos_bbox_func(params=params)
    inst_unit_size = [inst_xy[1, 0] - inst_xy[0, 0], inst_xy[1, 1] - inst_xy[0, 1]]

    # Generate and return the final instance
    inst = laygo2.object.VirtualInstance(
        name=name,
        xy=np.array([0, 0]),
        libname="mylib",
        cellname="myvcell_" + devtype,
        native_elements=nelements,
        shape=shape,
        pitch=pitch,
        transform=transform,
        unit_size=inst_unit_size,
        pins=pins,
    )
    return inst


def pmos_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return mos_generate_func(devtype="pmos", name=name, shape=shape, pitch=pitch, transform=transform, params=params)


def nmos_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return mos_generate_func(devtype="nmos", name=name, shape=shape, pitch=pitch, transform=transform, params=params)


# Spacers
def space_bbox_func(params):
    """Computes x and y coordinate values from params."""
    us = np.array([30, 100])  # unit space
    xy0 = [0, 0]
    xy1 = [us[0] * params["nf"], us[1]]
    xy = np.array([xy0, xy1])
    return xy


def space_pins_func(params):
    return None


def space_generate_func(devtype="pmos", name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    # Compute parameters
    nf = params["nf"] if "nf" in params.keys() else 1
    us = np.array([30, 100])  # unit space
    
    # Create the base mosfet structure.
    nelements = dict()
    # prboundary
    rxy = np.array([[0, 0], [us[0] * nf, us[1]]])
    nelements["PRBND0"] = laygo2.object.Rect(xy=rxy, layer=["prBoundary", "drawing"], name="PRBND0")
    # well
    rxy = np.array([[-20, 0], [us[0] * nf + 20, us[1]]])
    nelements["NW0"] = laygo2.object.Rect(xy=rxy, layer=["nwell", "drawing"], name="NW0")
    # implant
    rxy = np.array([[0, 5], [us[0] * nf, us[1] - 5]])
    if devtype == "nmos":
        nelements["NIMPL0"] = laygo2.object.Rect(xy=rxy, layer=["nimplant", "drawing"], name="NIMPL0")
    else:
        nelements["PIMPL0"] = laygo2.object.Rect(xy=rxy, layer=["pimplant", "drawing"], name="PIMPL0")
    # metal2 - rail
    rxy = np.array([[0, -5], [us[0] * nf, 5]])
    nelements["M2RAIL0"] = laygo2.object.Rect(xy=rxy, layer=["metal2", "drawing"], name="M2RAIL0")
    
    # Unit size
    inst_xy = space_bbox_func(params=params)
    inst_unit_size = [inst_xy[1, 0] - inst_xy[0, 0], inst_xy[1, 1] - inst_xy[0, 1]]
    
    # Generate and return the final instance
    inst = laygo2.object.VirtualInstance(
        name=name,
        xy=np.array([0, 0]),
        libname="mylib",
        cellname="myvcell_space_" + devtype,
        native_elements=nelements,
        shape=shape,
        pitch=pitch,
        transform=transform,
        unit_size=inst_unit_size,
        pins=None,
    )
    return inst


def pspace_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return space_generate_func(devtype="pmos", name=name, shape=shape, pitch=pitch, transform=transform, params=params)


def nspace_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return space_generate_func(devtype="nmos", name=name, shape=shape, pitch=pitch, transform=transform, params=params)


# Vias
def via_generate_func(devtype, name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    """Generates an instance from the input parameters."""
    # Create the via structure.
    nelements = dict()
    if devtype.startswith("via_r12"):
        layer = ["via1", "drawing"]
        xy = [[-5, -5], [5, 5]]
    if devtype.startswith("via_r23"):
        layer = ["via2", "drawing"]
        xy = [[-5, -5], [5, 5]]
    if devtype.startswith("via_r34"):
        layer = ["via3", "drawing"]
        xy = [[-5, -5], [5, 5]]
    nelements["VIA0"] = laygo2.object.Rect(xy=xy, layer=layer, name="VIA0")

    # Generate and return the final instance
    inst = laygo2.object.VirtualInstance(
        name=name,
        xy=np.array([0, 0]),
        libname="mylib",
        cellname="myvcell_devtype",
        native_elements=nelements,
        shape=shape,
        pitch=pitch,
        transform=transform,
        unit_size=np.array([0, 0]),
        pins=None,
    )
    return inst


def via_r12_default_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return via_generate_func(
        devtype="via_r12_default", name=name, shape=shape, pitch=pitch, transform=transform, params=params
    )


def via_r23_default_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return via_generate_func(
        devtype="via_r23_default", name=name, shape=shape, pitch=pitch, transform=transform, params=params
    )

def via_r34_default_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None):
    return via_generate_func(
        devtype="via_r34_default", name=name, shape=shape, pitch=pitch, transform=transform, params=params
    )


# Template library
def load_templates():
    """Load template to a template library object"""
    tlib = laygo2.object.database.TemplateLibrary(name="advtech_templates")
    # Transistors
    # Transistor layouts are created in laygo and stored as a virtual instance.
    tnmos = laygo2.object.template.UserDefinedTemplate(
        name="nmos", bbox_func=mos_bbox_func, pins_func=nmos_pins_func, generate_func=nmos_generate_func
    )
    tpmos = laygo2.object.template.UserDefinedTemplate(
        name="pmos", bbox_func=mos_bbox_func, pins_func=pmos_pins_func, generate_func=pmos_generate_func
    )
    tlib.append([tpmos, tnmos])
    # Spacers
    tnspace = laygo2.object.template.UserDefinedTemplate(
        name="nspace", bbox_func=space_bbox_func, pins_func=space_pins_func, generate_func=nspace_generate_func
    )
    tpspace = laygo2.object.template.UserDefinedTemplate(
        name="pspace", bbox_func=space_bbox_func, pins_func=space_pins_func, generate_func=pspace_generate_func
    )
    tlib.append([tpspace, tnspace])
    # Vias
    # Via layouts are created in laygo and stored as a virtual instance.
    tvia_r12_default = laygo2.object.template.UserDefinedTemplate(
        name="via_r12_default",
        bbox_func=lambda params: np.array([0, 0]),
        pins_func=lambda params: None,
        generate_func=via_r12_default_generate_func,
    )
    tlib.append([tvia_r12_default])
    tvia_r23_default = laygo2.object.template.UserDefinedTemplate(
        name="via_r23_default",
        bbox_func=lambda params: np.array([0, 0]),
        pins_func=lambda params: None,
        generate_func=via_r23_default_generate_func,
    )
    tlib.append([tvia_r23_default])
    tvia_r34_default = laygo2.object.template.UserDefinedTemplate(
        name="via_r34_default",
        bbox_func=lambda params: np.array([0, 0]),
        pins_func=lambda params: None,
        generate_func=via_r34_default_generate_func,
    )
    tlib.append([tvia_r34_default])
    return tlib


# Tests
if __name__ == "__main__":
    # Create templates.
    print("Create templates")
    templates = load_templates()
    print(templates["nmos"])
    print(templates["pmos"])



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interface with virtuoso in skill language

"""
import logging
from math import log10
from decimal import *

import numpy as np
import laygo2.object
import laygo2.util.transform as tf

__author__ = ""
__maintainer__ = ""
__status__ = "Prototype"


def _py2skill_number(value, scale=0.001):
    fmt_str = "%." + "%d" % (-1 * log10(scale) + 1) + "f "  # for truncations
    return fmt_str % (value * scale)


def _py2skill_float(value, scale=0.001):
    exp = np.ceil(-1 * log10(value) + 1)
    fmt_str = "%." + "%d" % (-1 * log10(scale) + 1) + "f"  # for truncations
    # convert to exponentail expression
    print(value, exp, str(value))
    return fmt_str % (value * (10**exp) * scale) + "e" + str(int(-1 * exp))


def _py2skill_list(pylist, scale=0.001):
    """Convert a python list object to a skill list."""
    list_str = "list( "
    for item in pylist:
        if isinstance(item, list):  # nested list
            list_str += _py2skill_list(item) + " "
        elif isinstance(item, np.ndarray):  # nested list
            list_str += _py2skill_list(item, scale=scale) + " "
        elif isinstance(item, str):
            list_str += '"' + str(item) + '" '
        elif isinstance(item, bool):
            if item:
                list_str += "t "
            else:
                list_str += "nil "
        elif isinstance(item, int) or isinstance(item, np.integer):
            # fmt_str = "%."+"%d" % (-1*log10(scale)+1)+"f "  # for truncations
            # list_str += fmt_str%(item*scale) + " "
            list_str += _py2skill_number(item, scale) + " "
    list_str += ")"
    return list_str


def _py2skill_inst_params_list(pylist):
    """Convert a python list object to a skill list. (for pcell parameters)"""
    list_str = "list( "
    for item in pylist:
        if isinstance(item, list):  # nested list
            list_str += _py2skill_inst_params_list(item) + " "
        elif isinstance(item, np.ndarray):  # nested list
            list_str += _py2skill_inst_params_list(item) + " "
        elif isinstance(item, str):
            list_str += '"' + str(item) + '" '
        elif isinstance(item, float):
            list_str += str(item)
            # list_str += _py2skill_float(item, scale) + " "
        elif isinstance(item, bool):
            if item:
                list_str += "t "
            else:
                list_str += "nil "
        elif isinstance(item, int) or isinstance(item, np.integer):
            # fmt_str = "%."+"%d" % (-1*log10(scale)+1)+"f "  # for truncations
            # list_str += fmt_str%(item*scale) + " "
            # list_str += _py2skill_number(item, 1) + " "  # do not scale integers
            list_str += str(item) + " "  # do not scale integers
    list_str += ")"
    return list_str


def _py2skill_inst_params(value_dict):
    """Convert instance parameter dictionary to skill list"""
    _list = []
    for k, v in value_dict.items():
        if isinstance(v, str):
            _type = "string"
        elif isinstance(v, float):
            _type = "float"
        elif isinstance(v, bool):
            _type = "boolean"
        elif isinstance(v, int) or isinstance(v, np.integer):
            _type = "int"
        _list.append([k, _type, v])
    return _py2skill_inst_params_list(_list)


def _translate_obj(objname, obj, scale=0.001, master=None, offset=np.array([0, 0])):
    """
    Convert an object to corresponding skill commands.
    offset : np.array([int, int])
        Offsets to obj.xy
    """
    if master is None:
        mxy = np.array([0, 0])
        mtf = "R0"
    else:  # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform
    if obj.__class__ == laygo2.object.Rect:
        # coloring
        if obj.color == None:
            color = "not MPT"
        else:
            color = obj.color
        # Invoke _laygo2_generate_rect( cv layer bbox ) in {header_filename}
        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(
            _xy
            + np.array(
                [[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]
            ),
            tf.Mt(mtf).T,
        )
        # _xy = mxy + np.dot(obj.xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
        #                   tf.Mt(mtf).T)
        return '_laygo2_generate_rect(cv, %s, %s, "%s") ; for the Rect object %s \n' % (
            _py2skill_list(obj.layer),
            _py2skill_list(_xy, scale=scale),
            color,
            objname,
        )  # coloring func. added
    elif obj.__class__ == laygo2.object.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        for idx, _obj in np.ndenumerate(_objelem):
            # Invoke _laygo2_generate_pin(cv, name, layer, bbox) in {header_filename}
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            return (
                '_laygo2_generate_pin(cv, "%s", %s, %s ) ; for the Pin object %s \n'
                % (
                    _obj.netname,
                    _py2skill_list(_obj.layer),
                    _py2skill_list(_xy, scale=scale),
                    objname,
                )
            )
    elif obj.__class__ == laygo2.object.Text:
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.Instance:
        # Invoke _laygo2_generate_instance( cv name libname cellname viewname loc orient num_rows num_cols
        # sp_rows sp_cols params params_order )
        _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
        if master is None:
            transform = obj.transform
        else:  # if the translated object has a master (e.g. VirtualInstance)
            transform = tf.combine(obj.transform, master.transform)
        if obj.shape is None:
            num_rows = 1
            num_cols = 1
            sp_rows = 0
            sp_cols = 0
        else:
            num_rows = obj.shape[1]
            num_cols = obj.shape[0]
            sp_rows = _py2skill_number(obj.pitch[1])
            sp_cols = _py2skill_number(obj.pitch[0])
        if obj.params is None:
            inst_params = "nil"
        else:
            inst_params = _py2skill_inst_params(obj.params["pcell_params"])
            # inst_params = _py2skill_list([["Wfg", "string", "500n"], ["fingers", "string", "4"], ["l", "string", "500n"]])
        return (
            '_laygo2_generate_instance(cv, "%s", "%s", "%s", "%s", %s, "%s", %d, %d, %s, %s, %s, %s) '
            "; for the Instance object %s \n"
            % (
                objname,
                obj.libname,
                obj.cellname,
                obj.viewname,
                _py2skill_list(_xy),
                transform,
                num_rows,
                num_cols,
                sp_rows,
                sp_cols,
                inst_params,
                "nil",
                objname,
            )
        )
    elif obj.__class__ == laygo2.object.VirtualInstance:
        cmd = ""
        if obj.shape is None:
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.Pin:
                    if obj.name == None:
                        obj.name = "NoName"
                    else:
                        pass
                    cmd += _translate_obj(
                        obj.name + "_" + elem_name, elem, scale=scale, master=obj
                    )
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.Pin:
                        cmd += _translate_obj(
                            obj.name + "_" + elem_name + str(i) + "_" + str(j),
                            elem,
                            scale=scale,
                            master=obj[i, j],
                        )
        return cmd
    else:
        return obj.translate_to_skill()  #

    return ""


def export(
    db, filename=None, cellname=None, scale=1e-3, reset_library=False, tech_library=None
):
    """
    Export a laygo2.object.database.Library object to skill code.

    Parameters
    ----------
    db: laygo2.database.Library
        The library database to exported.
    filename: str, optional
        If specified, the generated skill script is stored in filename.
    cellname: str or List[str]
        The name(s) of cell(s) to be exported.
    scale: float
        The scaling factor between laygo2's integer coordinats actual physical coordinates.
    reset_library: bool, optional
        If True, the library to export the cells is reset.
    tech_library: str, optional
        The name of technology library to be attached to the resetted library.

    Returns
    -------
    str: The generated skill script.

    Example
    --------
    >>> import laygo2
    >>> from laygo2.object.database import Design
    >>> from laygo2.object.physical import Rect, Pin, Instance, Text
    >>> # Create a design.
    >>> dsn = Design(name="mycell", libname="genlib")
    >>> # Create layout objects.
    >>> r0 = Rect(xy=[[0, 0], [100, 100]], layer=["M1", "drawing"])
    >>> p0 = Pin(xy=[[0, 0], [50, 50]], layer=["M1", "pin"], name="P")
    >>> i0 = Instance(libname="tlib", cellname="t0", name="I0", xy=[0, 0])
    >>> t0 = Text(xy=[[50, 50], [100, 100]], layer=["text", "drawing"], text="T")
    >>> # Add the layout objects to the design object.
    >>> dsn.append(r0)
    >>> dsn.append(p0)
    >>> dsn.append(i0)
    >>> dsn.append(t0)
    >>> #
    >>> # Export to skill.
    >>> lib = laygo2.object.database.Library(name="mylib")
    >>> lib.append(dsn)
    >>> scr = laygo2.interface.skill.export(lib, filename="myscript.il")
    >>> print(scr)
    ; (definitions of laygo2 skill functions)
    ; exporting mylib__mycell
    cv = _laygo2_open_layout("mylib" "mycell" "layout")
    _laygo2_generate_rect(cv, list( "M1" "drawing" ), list( list( 0.0000  0.0000  ) list( 0.1000  0.1000  ) ), "None")
    _laygo2_generate_pin(cv, "P", list( "M1" "pin" ), list( list( 0.0000  0.0000  ) list( 0.0500  0.0500  ) ) )
    _laygo2_generate_instance(cv, "I0", "tlib", "t0", "layout", list( 0.0000  0.0000  ), "R0", 1, 1, 0, 0, nil, nil)
    _laygo2_save_and_close_layout(cv)

    """
    # parse header functions.
    cmd = "; laygo2 layout export skill script.\n\n"
    import os

    header_filename = (
        os.path.abspath(laygo2.interface.__file__)[:-11] + "skill_export.il"
    )
    with open(header_filename, "r") as f:
        cmd += f.read()
        cmd += "\n"

    cellname = (
        db.keys() if cellname is None else cellname
    )  # export all cells if cellname is not given.
    cellname = (
        [cellname] if isinstance(cellname, str) else cellname
    )  # convert to a list for iteration.
    if reset_library:
        cmd += '_laygo2_reset_library("%s" "%s")\n' % (db.name, tech_library)
    for cn in cellname:
        cmd += "\n; exporting %s__%s\n" % (db.name, cn)  # open the design.
        logging.debug("Export_to_SKILL: Cellname:" + cn)
        # cmd += "let( (cv) \n"  # limit the scope of cv. Disabled to prevent overflow.
        cmd += 'cv = _laygo2_open_layout("%s" "%s" "layout")\n' % (
            db.name,
            cn,
        )  # open the design.
        # export objects
        for objname, obj in db[cn].items():
            cmd += _translate_obj(objname, obj, scale=scale)
        cmd += "_laygo2_save_and_close_layout(cv)\n"  # close the layout
        # cmd += ");let\n" Disabled to prevent overflow.
    if filename is not None:  # export to a file.
        with open(filename, "w") as f:
            f.write(cmd)
    return cmd



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
**laygo2.object.template** module implements classes for various layout template objects.
"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

from abc import *
import numpy as np
import laygo2.object


class Template(metaclass=ABCMeta):
    """
    The base class that defines the functions and attributes of the template.

    Notes
    -----
    **(Korean)** Template의 기본동작과 속성을 정의한 기본 클래스.
    """

    name = None
    """str: Template name."""

    def __init__(self, name=None):
        """
        Constructor function of Template class.

        Parameters
        ----------
        name : str or None, optional.
            The name of this template.
        """
        self.name = name

    def __str__(self):
        """Return a string corresponding to this object's information."""
        return self.summarize()

    def summarize(self):
        """Return the summary of the template information."""
        return (
            self.__repr__() + " "
            "name: "
            + self.name
            + ", "
            + "class: "
            + self.__class__.__name__
            + ", "
            + ""
        )

    def height(self, params=None):
        """int: Return the height of a template."""
        return abs(self.bbox(params=params)[0, 1] - self.bbox(params=params)[1, 1])

    def width(self, params=None):
        """int: Return the width of a template."""
        return abs(self.bbox(params=params)[0, 0] - self.bbox(params=params)[1, 0])

    def size(self, params=None):
        """int: Return the size of a template."""
        return np.array([self.width(params=params), self.height(params=params)])

    @abstractmethod
    def bbox(self, params=None):
        """numpy.ndarray: (Abstract method) Return the bounding box of a template."""
        pass

    @abstractmethod
    def pins(self, params=None):
        """dict: (Abstract method) Return dict having the collection of pins of a template."""
        pass

    @abstractmethod
    def generate(self, name=None, shape=None, pitch=None, transform="R0", netmap=None, params=None):
        """instance: (Abstract method) Return the instance generated from a template."""
        pass
        


class NativeInstanceTemplate(Template):
    """
    NativeInstanceTemplate class implements the template that generate Instance.

    Notes
    -----
    **(Korean)** NativeInstanceTemplate 클래스는 기본 Instance를 반환하는 템플릿을 구현한다.
    """

    libname = None
    """str: Library name of NativeInstanceTemplate object.

    Example
    -------
    >>> from laygo2.object.physical import Pin
    >>> from laygo2.object.template import NativeInstanceTemplate
    >>> p = dict() 
    >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                     netname='i')
    >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                     netname='o')
    >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                    bbox=[[0, 0], [100, 100]], pins=p)
    >>> nt.libname 
    'mylib'
    
    .. image:: ../assets/img/object_template_NativeInstanceTemplate_libname.png
          :height: 250

    Notes
    -----
    **(Korean)** NativeInstanceTemplate 객체의 library 이름.
    """

    cellname = None
    """str: Cellname of NativeInstanceTemplate object.

    Example
    -------
    >>> from laygo2.object.physical import Pin
    >>> from laygo2.object.template import NativeInstanceTemplate
    >>> p = dict() 
    >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                     netname='i')
    >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                     netname='o')
    >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                    bbox=[[0, 0], [100, 100]], pins=p)
    >>> nt.cellname 
    'mytemp'
    
    .. image:: ../assets/img/object_template_NativeInstanceTemplate_cellname.png
          :height: 250

    Notes
    -----
    **(Korean)** NativeInstanceTemplate 객체의 cellname.
    """

    _bbox = np.array([[0, 0], [0, 0]])

    _pins = None

    def __init__(self, libname, cellname, bbox=np.array([[0, 0], [0, 0]]), pins=None):
        """
        Constructor function of NativeInstanceTemplate class.

        Parameters
        ----------
        libname : str
            library name.
        cellname : str
            cell name.
        bbox : numpy.ndarray
            bbox.
        pins : dict
            dictionary having the pin object.

        Returns
        -------
        NativeInstanceTemplate

        Example
        -------
        >>> from laygo2.object.physical import Pin
        >>> from laygo2.object.template import NativeInstanceTemplate
        >>> p = dict() 
        >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                         netname='i')
        >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                         netname='o')
        >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                        bbox=[[0, 0], [100, 100]], pins=p)
        <laygo2.object.template.NativeInstanceTemplate object>
        >>> print(nt)
        <laygo2.object.template.NativeInstanceTemplate object at 0x0000013C01..> 
         name: mytemp, class: NativeInstanceTemplate, bbox: [[0, 0], [100, 100]], 
         pins: {'i': <laygo2.object.physical.Pin object at 0x0000013C01CEFDC0>, 
                'o': <laygo2.object.physical.Pin object at 0x0000013C01C30BE0>},

        .. image:: ../assets/img/object_template_NativeInstanceTemplate_init.png
          :height: 250

        Notes
        -----
        **(Korean)** NativeInstanceTemplate 클래스의 생성자함수.
        """
        self.libname = libname
        self.cellname = cellname
        self._bbox = None if bbox is None else np.asarray(bbox)
        self._pins = pins
        Template.__init__(self, name=cellname)

    def summarize(self):
        """Return the summary of the template information."""
        return (
            self.__repr__() + " "
            "name: "
            + self.name
            + ", "
            + "class: "
            + self.__class__.__name__
            + ", "
            + "bbox: "
            + str(self.bbox().tolist())
            + ", "
            + "pins: "
            + str(self.pins())
            + ", "
            + ""
        )

    # Core template functions
    def bbox(self, params=None):
        """
        bbox of NativeInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray

        Example
        -------
        >>> from laygo2.object.physical import Pin
        >>> from laygo2.object.template import NativeInstanceTemplate
        >>> p = dict() 
        >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                         netname='i')
        >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                         netname='o')
        >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                        bbox=[[0, 0], [100, 100]], pins=p)
        >>> nt.bbox()
        array([[  0,   0],
               [100, 100]])

        .. image:: ../assets/img/object_template_NativeInstanceTemplate_bbox.png
          :height: 250

        Notes
        -----
        **(Korean)** NativeInstanceTemplate 객체의 bbox.
        """
        return self._bbox

    def pins(self, params=None):
        """
        Return pin dictionary of NativeInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        dict : A dictionary that contains pin information.

        Example
        -------
        >>> from laygo2.object.physical import Pin
        >>> from laygo2.object.template import NativeInstanceTemplate
        >>> p = dict() 
        >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                         netname='i')
        >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                         netname='o')
        >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                        bbox=[[0, 0], [100, 100]], pins=p)
        >>> nt.pins()
        >>> {'i': <laygo2.object.physical.Pin object at 0x000002578C762500>, 
             'o': <laygo2.object.physical.Pin object at 0x00000257FBA87C40>}                   
        >>> print(nt.pins()['i']) 
            <laygo2.object.physical.Pin object at 0x000002578C762500> 
                name: None,
                class: Pin,
                xy: [[0, 0], [10, 10]],
                params: None, 
                layer: ['M1' 'drawing'], 
                netname: i, 
                shape: None, 
                master: None

        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_pins.png
          :height: 250

        Notes
        -----
        **(Korean)** NativeInstanceTemplate 객체의 pin dictionary 반환.
        """
        return self._pins

    def generate(self, name=None, shape=None, pitch=None, transform="R0", netmap=None, params=None):
        """
        Generate Instance object.

        Parameters
        ----------
        name : str
            name of the instance to be generated.
        shape : numpy.ndarray, optional.
            shape of the object to be generated.
        pitch : numpy.ndarray, optional.
            pitch of the object to be generated.
        transform : str
            transformation attribute of the object to be generated.
        netmap : dict, optional.
            dictionary containing netmap conversion information of pins.
        params : dict, optional.
            dictionary having the object attributes.

        Returns
        -------
        (laygo2.object.physical.Instance) generated Instance object

        See Also
        --------
        Class Instance

        Example
        -------
        >>> from laygo2.object.physical import Pin
        >>> from laygo2.object.template import NativeInstanceTemplate
        >>> p = dict() 
        >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                         netname='i')
        >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                         netname='o')
        >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                        bbox=[[0, 0], [100, 100]], pins=p)
        >>> i0 = nt.generate(name="I0")
        >>> i1 = nt.generate(name="I1")
        >>> print(i0)
        <laygo2.object.physical.Instance object at 0x000002C3F717F850> 
        name: I0,
        class: Instance,
        xy: [0, 0],
        params: None,
        size: [100, 100],
        shape: None,
        pitch: [100, 100],
        transform: R0,
        pins: {'i': <laygo2.object.physical.Pin object at 0x000002C3F717F820>, 
               'o': <laygo2.object.physical.Pin object at 0x000002C3F717FD90>},
        >>> print(i1)
        <laygo2.object.physical.Instance object at 0x000002C3FF91C100> 
        name: I1,
        class: Instance,
        xy: [0, 0],
        params: None,
        size: [100, 100],
        shape: None,
        pitch: [100, 100],
        transform: R0,
        pins: {'i': <laygo2.object.physical.Pin object at 0x000002C3FF91C0A0>, 
               'o': <laygo2.object.physical.Pin object at 0x000002C3FF91C070>},
        
        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_generate.png
          :height: 250

        Notes
        -----
        **(Korean)** 템플릿으로부터 Instance 객체 생성.
        """
        inst = laygo2.object.physical.Instance(
            libname=self.libname,
            cellname=self.cellname,
            xy=np.array([0, 0]),
            shape=shape,
            pitch=pitch,
            unit_size=self.size(params),
            pins=self.pins(params),
            transform=transform,
            name=name,
            params=params,
        )
        # update netnames if netmap is provided.
        if netmap is not None:
            inst.update_netname(netmap=netmap)  
        return inst 

    # I/O functions
    def export_to_dict(self):
        """
        Return Dictionary containing the information of NativeInstanceTemplate.

        Parameters
        ----------
        None

        Returns
        -------
        dict

        See Also
        --------
        Class Instance

        Example
        -------
        >>> from laygo2.object.physical import Pin
        >>> from laygo2.object.template import NativeInstanceTemplate
        >>> p = dict() 
        >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], 
                         netname='i')
        >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], 
                         netname='o')
        >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp', 
                                        bbox=[[0, 0], [100, 100]], pins=p)
        >>> nt.export_to_dict()
            {'libname': 'mylib', 
             'cellname': 'mytemp', 
             'bbox': [[0, 0], [100, 100]], 
             'pins': {'i': {'xy': [[0, 0], [10, 10]], 
                            'layer': ['M1', 'drawing'], 
                            'name': None, 
                            'netname': 'i'}, 
                      'o': {'xy': [[90, 90], [100, 100]], 
                            'layer': ['M1', 'drawing'], 
                            'name': None, 
                            'netname': 'o'}
                     }
            }

        .. image:: ../assets/img/object_template_NativeInstanceTemplate_export_to_dict.png
          :height: 250

        Notes
        -----
        **(Korean)** NativeInstanceTemplate 의 정보가 담긴 Dictonary 반환.
        """
        db = dict()
        db["libname"] = self.libname
        db["cellname"] = self.cellname
        db["bbox"] = self.bbox().tolist()
        db["pins"] = dict()
        for pn, p in self.pins().items():
            db["pins"][pn] = p.export_to_dict()
        return db


class ParameterizedInstanceTemplate(Template):
    """
    ParameterizedInstanceTemplate class implements the template that generates 
    an instance with variable size (bbox) and pin parameters, based on its 
    input parameters (such as instances mapped to Cadence Virtuoso's pcells 
    or Pycells). 

    Notes
    -----
    **(Korean)** ParameterizedInstanceTemplate 클래스는 입력 파라미터에 따라 
    크기와 pin 정보가 가변하는 인스턴스를 생성하는 템플릿을 구현한다. 
    """

    libname = None
    """str: The libname of the instance being generated."""

    cellname = None
    """str: The cellname of the instance being generated."""

    _bbox = None

    _pins = None

    def __init__(self, libname, cellname, bbox_func=None, pins_func=None):
        """
        Constructor of ParameterizedInstanceTemplate class.

        Parameters
        ----------
        libname : str
            The library name.
        cellname : str
            The cell name of the template.
        bbox_func : callable
            The function that computes the bounding box of the template from 
            its input parameters.
        pins_func : callable
            The function that returns a dictionary that contains its pin 
            objects for its input parameters.

        Returns
        -------
        laygo2.object.template.ParameterizedInstanceTemplate

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import ParameterizedInstanceTemplate
        >>> from laygo2.object.physical import Pin
        >>> # bbox computation function.
        >>> def pcell_bbox_func(params):
                return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def pcell_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # Create a template
        >>> pcell_temp = ParameterizedInstanceTemplate(
        >>>     libname="mylib",
        >>>     cellname="mypcelltemplate",
        >>>     bbox_func=pcell_bbox_func,
        >>>     pins_func=pcell_pins_func,
        >>> )
        >>> # Generate an instance for input parameters.
        >>> pcell_inst_params = {"mult": 4}
        >>> pcell_inst = pcell_temp.generate(
        >>>    name="mypcellinst",
        >>>    transform="R0",
        >>>    params=pcell_inst_params,
        >>> )
        >>> # display
        >>> print(pcell_temp)
            <laygo2.object.template.ParameterizedInstanceTemplate object at 0x000001B1BA91D9C0>
            name: mypcelltemplate, 
            class: ParameterizedInstanceTemplate,
        >>> print(pcell_inst)
            xy: [0, 0],
            params: {'mult': 4},
            size: [400, 100],
            shape: None,
            pitch: [400, 100],
            transform: R0,
            pins: {'in0': <laygo2.object.physical.Pin object at 0x000001B1BA91FCA0>, 
                   'out0': <laygo2.object.physical.Pin object at 0x000001B1BA91FC70>, 
                   'in1': <laygo2.object.physical.Pin object at 0x000001B1BA91FC10>, 
                   'out1': <laygo2.object.physical.Pin object at 0x000001B1BA91E8C0>, 
                   'in2': <laygo2.object.physical.Pin object at 0x000001B1BA91E890>, 
                   'out2': <laygo2.object.physical.Pin object at 0x000001B1BA91FBE0>, 
                   'in3': <laygo2.object.physical.Pin object at 0x000001B1BA91E7D0>, 
                   'out3': <laygo2.object.physical.Pin object at 0x000001B1BA91E710>},

        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_init.png
          :height: 250

        Notes
        -----
        **(Korean)** ParameterizedInstanceTemplate 클래스의 생성자함수.
        """
        self.libname = libname
        self.cellname = cellname
        self._bbox = bbox_func
        self._pins = pins_func
        Template.__init__(self, name=cellname)

    # Core template functions
    def bbox(self, params=None):
        """
        bbox of ParameterizedInstanceTemplate object.

        Parameters
        ----------
        params: dict
            A dictionary that contains input parameters corresponding to the 
            bounding box to be computed.

        Returns
        -------
        numpy.ndarray: A 2x2 numpy array that contains the bounding box 
            coordinates corresponding to the input parameters.

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import ParameterizedInstanceTemplate
        >>> from laygo2.object.physical import Pin
        >>> # bbox computation function.
        >>> def pcell_bbox_func(params):
                return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def pcell_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # Create a template
        >>> pcell_temp = ParameterizedInstanceTemplate(
        >>>     libname="mylib",
        >>>     cellname="mypcelltemplate",
        >>>     bbox_func=pcell_bbox_func,
        >>>     pins_func=pcell_pins_func,
        >>> )
        >>> # Compute bbox for input parameters
        >>> pcell_inst_params = {"mult": 4}
        >>> pcell_temp.bbox(params=pcell_inst_params)
        array([[  0,   0],
               [400, 100]])

        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_bbox.png
          :height: 250

        Notes
        -----
        **(Korean)** ParameterizedInstanceTemplate 객체의 bbox.
        """
        return self._bbox(params=params)

    def pins(self, params=None):
        """
        Return pin dictionary of ParameterizedInstanceTemplate object.

        Parameters
        ----------
        params: dict
            A dictionary that contains input parameters corresponding to the 
            pin objects to be produced.

        Returns
        -------
        dict: A dictionary that contains pin object corresponding to the 
            input parameters.

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import ParameterizedInstanceTemplate
        >>> from laygo2.object.physical import Pin
        >>> # bbox computation function.
        >>> def pcell_bbox_func(params):
                return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def pcell_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # Create a template
        >>> pcell_temp = ParameterizedInstanceTemplate(
        >>>     libname="mylib",
        >>>     cellname="mypcelltemplate",
        >>>     bbox_func=pcell_bbox_func,
        >>>     pins_func=pcell_pins_func,
        >>> )
        >>> # Compute bbox for input parameters
        >>> pcell_inst_params = {"mult": 4}
        >>> pcell_temp.pins(params=pcell_inst_params)
            {'in0': <laygo2.object.physical.Pin object at 0x000001B1BA91F7C0>, 
             'out0': <laygo2.object.physical.Pin object at 0x000001B1BA91E830>,
             'in1': <laygo2.object.physical.Pin object at 0x000001B1BA91DAB0>, 
             'out1': <laygo2.object.physical.Pin object at 0x000001B1BA91E860>, 
             'in2': <laygo2.object.physical.Pin object at 0x000001B1BA91E560>, 
             'out2': <laygo2.object.physical.Pin object at 0x000001B1BA91E800>, 
             'in3': <laygo2.object.physical.Pin object at 0x000001B1BA91E770>, 
             'out3': <laygo2.object.physical.Pin object at 0x000001B1BA91EA40>}

        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_pins.png
          :height: 250

        Notes
        -----
        **(Korean)** ParameterizedInstanceTemplate 객체의 pin dictionary 반환.
        """
        return self._pins(params=params)

    def generate(self, name=None, shape=None, pitch=None, transform="R0", netmap=None, params=None):
        """
        Generate an Instance object corresponding to the template and its 
        input parameters.

        Parameters
        ----------
        name : str
            name of the instance to be generated.
        shape : numpy.ndarray, optional.
            shape of the object to be generated.
        pitch : numpy.ndarray, optional.
            pitch of the object to be generated.
        transform : str, optional.
            transformation attribute of the entity to be generated.
        netmap : dict, optional.
            dictionary containing netmap conversion information of pins.
        params : dict, optional.
            dictionary having the entity attributes.

        Returns
        -------
        laygo2.object.physical.Instance: The generated Instance object

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import ParameterizedInstanceTemplate
        >>> from laygo2.object.physical import Pin
        >>> # bbox computation function.
        >>> def pcell_bbox_func(params):
                return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def pcell_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # Create a template
        >>> pcell_temp = ParameterizedInstanceTemplate(
        >>>     libname="mylib",
        >>>     cellname="mypcelltemplate",
        >>>     bbox_func=pcell_bbox_func,
        >>>     pins_func=pcell_pins_func,
        >>> )
        >>> # Generate an instance for input parameters.
        >>> pcell_inst_params = {"mult": 4}
        >>> pcell_inst = pcell_temp.generate(
        >>>    name="mypcellinst",
        >>>    transform="R0",
        >>>    params=pcell_inst_params,
        >>> )
        >>> # display
        >>> print(pcell_temp)
            <laygo2.object.template.ParameterizedInstanceTemplate object at 0x000001B1BA91D9C0>
            name: mypcelltemplate, 
            class: ParameterizedInstanceTemplate,
        >>> print(pcell_inst)
            xy: [0, 0],
            params: {'mult': 4},
            size: [400, 100],
            shape: None,
            pitch: [400, 100],
            transform: R0,
            pins: {'in0': <laygo2.object.physical.Pin object at 0x000001B1BA91FCA0>, 
                   'out0': <laygo2.object.physical.Pin object at 0x000001B1BA91FC70>, 
                   'in1': <laygo2.object.physical.Pin object at 0x000001B1BA91FC10>, 
                   'out1': <laygo2.object.physical.Pin object at 0x000001B1BA91E8C0>, 
                   'in2': <laygo2.object.physical.Pin object at 0x000001B1BA91E890>, 
                   'out2': <laygo2.object.physical.Pin object at 0x000001B1BA91FBE0>, 
                   'in3': <laygo2.object.physical.Pin object at 0x000001B1BA91E7D0>, 
                   'out3': <laygo2.object.physical.Pin object at 0x000001B1BA91E710>},

        .. image:: ../assets/img/object_template_ParameterizedInstanceTemplate_generate.png
          :height: 250

        Notes
        -----
        **(Korean)** 템플릿으로부터 입력 파라미터에 따른 Instance 객체 생성.

        파라미터
            - name(str): 생성할 인스턴스의 이름.
            - shape(numpy.ndarray): (optional) 생성할 객체의 배열 shape.
            - pitch(numpy.ndarray): (optional) 생성할 객체 간의 간격.
            - params(dict) : 개체의 속성이 담긴 dictionary.
            - transform(str): 생성할 개체의 변환 속성.
        반환값
            - laygo2.Instance: 생성된 객체
        """
        # xy = xy + np.dot(self.xy(params)[0], tf.Mt(transform).T)
        inst = laygo2.object.physical.Instance(
            libname=self.libname,
            cellname=self.cellname,
            xy=np.array([0, 0]),
            shape=shape,
            pitch=pitch,
            unit_size=self.size(params),
            pins=self.pins(params),
            transform=transform,
            name=name,
            params=params,
        )
        # update netnames if netmap is provided.
        if netmap is not None:
            inst.update_netname(netmap=netmap)  
        return inst


class UserDefinedTemplate(Template):
    """
    UserDefinedTemplate class implements the template that generates 
    a VirtualInstance object corresponding to the template and input 
    parameters.

    Notes
    -----
    **(Korean)** UserDefinedTemplate 클래스는 VirtualInstance를 반환하는 
    템플릿을 구현한다.

    """

    _bbox = None
    """The internal pointer to the bbox computing function."""

    _pins = None
    """The internal pointer to the pin creation function."""

    _generate = None
    """The internal pointer to the instance generation function."""

    def __init__(self, bbox_func, pins_func, generate_func, name=None):
        """
        Constructor function of UserDefinedTemplate class.

        Parameters
        ----------
        bbox_func: callable
            The function that computes the bounding box of the template from 
            its input parameters.
        pins_func: callable
            The function that returns a dictionary that contains its pin 
            objects for its input parameters.
        generate_func: callable
            The function that returns a generated VirtualInstance object 
            for its input parameters.
        name : str
            The name of the template.

        Returns
        -------
        laygo2.object.template.UserDefinedTemplate

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import UserDefinedTemplate
        >>> from laygo2.object.physical import Pin, Rect, VirtualInstance
        >>> # bbox computation function.
        >>> def user_bbox_func(params):
        >>>     return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def user_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # instance generation function.
        >>> def user_generate_func(
        >>>     name=None, shape=None, pitch=np.array([0, 0]), transform="R0", 
        >>>     params=None):
        >>>     m = params["mult"]
        >>>     shape = np.array([1, 1]) if shape is None else np.asarray(shape)
        >>>     inst_pins = user_pins_func(params)
        >>>     inst_native_elements = dict()
        >>>     for i in range(m):
        >>>         ofst = i * 100
        >>>         inst_native_elements["R0_" + str(i)] = Rect(
        >>>             xy=[[ofst, 0], [ofst + 10, 10]], layer=["M1", "drawing"]
        >>>         )
        >>>         inst_native_elements["R1_" + str(i)] = Rect(
        >>>             xy=[[ofst + 90, 90], [ofst + 100, 100]], layer=["M1", "drawing"]
        >>>         )
        >>>     inst_native_elements["R2"] = Rect(
        >>>         xy=[[0, 0], [m * 100, 100]], layer=["prBoundary", "drawing"]
        >>>     )
        >>>     inst = VirtualInstance(
        >>>         name=name,
        >>>         libname="mylib",
        >>>         cellname="myvinst",
        >>>         xy=np.array([0, 0]),
        >>>         native_elements=inst_native_elements,
        >>>         shape=shape,
        >>>         pitch=pitch,
        >>>         unit_size=[m * 100, 100],
        >>>         pins=inst_pins,
        >>>         transform=transform,
        >>>         params=params,
        >>>     )
        >>>     return inst
        >>> # UserDefinedTemplate construction.
        >>> user_temp = UserDefinedTemplate(
        >>>     name="myusertemplate",
        >>>     bbox_func=user_bbox_func,
        >>>     pins_func=user_pins_func,
        >>>     generate_func=user_generate_func,
        >>> )
        >>> # VirtualInstance generation.
        >>> user_inst = user_temp.generate(name="myinst", params={"mult": 5})
        >>> # Display
        >>> print(user_temp)
            <laygo2.object.template.UserDefinedTemplate object at 0x00000192BF990130> 
            name: myusertemplate, class: UserDefinedTemplate,
        >>> print(user_inst)
            <laygo2.object.physical.VirtualInstance object at 0x00000192BF990280>
                name: myinst,
                class: VirtualInstance,
                xy: [0, 0],
                params: {'mult': 5},
                size: [500, 100],
                shape: [1, 1],
                pitch: [500, 100],
                transform: R0,
                pins: {'in0': <laygo2.object.physical.Pin object at 0x00000192BF9930D0>, 
                       'out0': <laygo2.object.physical.Pin object at 0x00000192BF9931C0>, 
                       'in1': <laygo2.object.physical.Pin object at 0x00000192BF993760>, 
                       'out1': <laygo2.object.physical.Pin object at 0x00000192BF9936A0>, 
                       'in2': <laygo2.object.physical.Pin object at 0x00000192BF993610>, 
                       'out2': <laygo2.object.physical.Pin object at 0x00000192BF9935B0>, 
                       'in3': <laygo2.object.physical.Pin object at 0x00000192BF9932E0>, 
                       'out3': <laygo2.object.physical.Pin object at 0x00000192BF9931F0>, 
                       'in4': <laygo2.object.physical.Pin object at 0x00000192BF993130>, 
                       'out4': <laygo2.object.physical.Pin object at 0x00000192BF9930A0>},
                native elements: {'R0_0': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_0': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_1': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_1': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_2': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_2': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_3': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_3': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_4': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_4': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R2': <laygo2.object.physical.Rect object at 0x000...>}
        >>> print(user_inst.bbox)
            [[  0   0]
             [500 100]]

        .. image:: ../assets/img/object_template_UserDefinedTemplate_init.png
          :height: 250

        Notes
        -----
        **(Korean)** UserDefinedTemplate 클래스의 생성자 함수.
        
        파라미터
            - bbox_func(callable): bbox를 연산하는 함수.
            - pins_func(callable): pins를 생성하는 함수.
            - generate_func(callable): VirtualInstance를 생성하는 함수.
            - name(str): 템플릿 이름
        반환값
            - laygo2.object.template.UserDefinedTemplate
        """
        self._bbox = bbox_func
        self._pins = pins_func
        self._generate = generate_func
        Template.__init__(self, name=name)

    # Core template functions
    def bbox(self, params=None):
        """
        Return bbox of UserDefinedTemplate object.

        Parameters
        ----------
        params: dict
            A dictionary that contains input parameters corresponding to the 
            bounding box to be computed.

        Returns
        -------
        numpy.ndarray: A 2x2 numpy array that contains the bounding box 
            coordinates corresponding to the input parameters.

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import UserDefinedTemplate
        >>> from laygo2.object.physical import Pin, Rect, VirtualInstance
        >>> # bbox computation function.
        >>> def user_bbox_func(params):
        >>>     return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def user_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # instance generation function.
        >>> def user_generate_func(
        >>>     name=None, shape=None, pitch=np.array([0, 0]), transform="R0", 
        >>>     params=None):
        >>>     m = params["mult"]
        >>>     shape = np.array([1, 1]) if shape is None else np.asarray(shape)
        >>>     inst_pins = user_pins_func(params)
        >>>     inst_native_elements = dict()
        >>>     for i in range(m):
        >>>         ofst = i * 100
        >>>         inst_native_elements["R0_" + str(i)] = Rect(
        >>>             xy=[[ofst, 0], [ofst + 10, 10]], layer=["M1", "drawing"]
        >>>         )
        >>>         inst_native_elements["R1_" + str(i)] = Rect(
        >>>             xy=[[ofst + 90, 90], [ofst + 100, 100]], layer=["M1", "drawing"]
        >>>         )
        >>>     inst_native_elements["R2"] = Rect(
        >>>         xy=[[0, 0], [m * 100, 100]], layer=["prBoundary", "drawing"]
        >>>     )
        >>>     inst = VirtualInstance(
        >>>         name=name,
        >>>         libname="mylib",
        >>>         cellname="myvinst",
        >>>         xy=np.array([0, 0]),
        >>>         native_elements=inst_native_elements,
        >>>         shape=shape,
        >>>         pitch=pitch,
        >>>         unit_size=[m * 100, 100],
        >>>         pins=inst_pins,
        >>>         transform=transform,
        >>>         params=params,
        >>>     )
        >>>     return inst
        >>> # UserDefinedTemplate construction.
        >>> user_temp = UserDefinedTemplate(
        >>>     name="myusertemplate",
        >>>     bbox_func=user_bbox_func,
        >>>     pins_func=user_pins_func,
        >>>     generate_func=user_generate_func,
        >>> )
        >>> user_temp.bbox(params={"mult": 5})
        array([[  0,   0],
               [500, 100]])

        .. image:: ../assets/img/object_template_UserDefinedTemplate_bbox.png
          :height: 250

        Notes
        -----
        **(Korean)** UserDefinedTemplate 객체의 bbox 반환.
        """
        return self._bbox(params=params)

    def pins(self, params=None):
        """
        Pins of UserDefinedTemplate object.

        Parameters
        ----------
        params: dict
            A dictionary that contains input parameters corresponding to the 
            pin objects to be produced.

        Returns
        -------
        dict: A dictionary that contains pin object corresponding to the 
            input parameters.

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import UserDefinedTemplate
        >>> from laygo2.object.physical import Pin, Rect, VirtualInstance
        >>> # bbox computation function.
        >>> def user_bbox_func(params):
        >>>     return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def user_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # instance generation function.
        >>> def user_generate_func(
        >>>     name=None, shape=None, pitch=np.array([0, 0]), transform="R0", 
        >>>     params=None):
        >>>     m = params["mult"]
        >>>     shape = np.array([1, 1]) if shape is None else np.asarray(shape)
        >>>     inst_pins = user_pins_func(params)
        >>>     inst_native_elements = dict()
        >>>     for i in range(m):
        >>>         ofst = i * 100
        >>>         inst_native_elements["R0_" + str(i)] = Rect(
        >>>             xy=[[ofst, 0], [ofst + 10, 10]], layer=["M1", "drawing"]
        >>>         )
        >>>         inst_native_elements["R1_" + str(i)] = Rect(
        >>>             xy=[[ofst + 90, 90], [ofst + 100, 100]], layer=["M1", "drawing"]
        >>>         )
        >>>     inst_native_elements["R2"] = Rect(
        >>>         xy=[[0, 0], [m * 100, 100]], layer=["prBoundary", "drawing"]
        >>>     )
        >>>     inst = VirtualInstance(
        >>>         name=name,
        >>>         libname="mylib",
        >>>         cellname="myvinst",
        >>>         xy=np.array([0, 0]),
        >>>         native_elements=inst_native_elements,
        >>>         shape=shape,
        >>>         pitch=pitch,
        >>>         unit_size=[m * 100, 100],
        >>>         pins=inst_pins,
        >>>         transform=transform,
        >>>         params=params,
        >>>     )
        >>>     return inst
        >>> # UserDefinedTemplate construction.
        >>> user_temp = UserDefinedTemplate(
        >>>     name="myusertemplate",
        >>>     bbox_func=user_bbox_func,
        >>>     pins_func=user_pins_func,
        >>>     generate_func=user_generate_func,
        >>> )
        >>> user_temp.pins(params={"mult": 5})
        {'in0': <laygo2.object.physical.Pin object at 0x00000192BF990670>, 
        'out0': <laygo2.object.physical.Pin object at 0x00000192BF990400>, 
        'in1': <laygo2.object.physical.Pin object at 0x00000192BF993250>, 
        'out1': <laygo2.object.physical.Pin object at 0x00000192BF9903D0>, 
        'in2': <laygo2.object.physical.Pin object at 0x00000192BF9901F0>, 
        'out2': <laygo2.object.physical.Pin object at 0x00000192BF9904F0>, 
        'in3': <laygo2.object.physical.Pin object at 0x00000192BF993640>, 
        'out3': <laygo2.object.physical.Pin object at 0x00000192BF990520>, 
        'in4': <laygo2.object.physical.Pin object at 0x00000192BF9936D0>, 
        'out4': <laygo2.object.physical.Pin object at 0x00000192BF993790>}

        .. image:: ../assets/img/object_template_UserDefinedTemplate_pins.png
          :height: 250

        Notes
        -----
        **(Korean)** UserDefinedTemplate 객체의 pin dictionary를 생성 및 반환하는 함수.
        """
        return self._pins(params=params)

    def generate(self, name=None, shape=None, pitch=None, transform="R0", netmap=None, params=None):
        """
        Generate a VirtualInstance object by calling generate_func() bound to the template.

        Parameters
        ----------
        name : str
            name of the instance to be generated.
        shape : numpy.ndarray, optional.
            shape of the object to be generated.
        pitch : numpy.ndarray, optional.
            pitch of the object to be generated.
        transform : str, optional.
            transformation attribute of the entity to be generated.
        netmap : dict, optional.
            dictionary containing netmap conversion information of pins.
        params : dict, optional.
            dictionary having the entity attributes.

        Returns
        -------
        laygo2.object.physical.VirtualInstance: The generated VirtualInstance object.

        Example
        -------
        >>> import numpy as np
        >>> from laygo2.object.template import UserDefinedTemplate
        >>> from laygo2.object.physical import Pin, Rect, VirtualInstance
        >>> # bbox computation function.
        >>> def user_bbox_func(params):
        >>>     return np.array([[0, 0], [100 * params["mult"], 100]])
        >>> # pin generation function.
        >>> def user_pins_func(params):
        >>>     template_pins = dict()
        >>>     for i in range(params["mult"]):
        >>>         template_pins["in" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="in" + str(i),
        >>>         )
        >>>         template_pins["out" + str(i)] = Pin(
        >>>             xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
        >>>             layer=["M1", "drawing"],
        >>>             netname="out" + str(i),
        >>>         )
        >>>     return template_pins
        >>> # instance generation function.
        >>> def user_generate_func(
        >>>     name=None, shape=None, pitch=np.array([0, 0]), transform="R0", 
        >>>     params=None):
        >>>     m = params["mult"]
        >>>     shape = np.array([1, 1]) if shape is None else np.asarray(shape)
        >>>     inst_pins = user_pins_func(params)
        >>>     inst_native_elements = dict()
        >>>     for i in range(m):
        >>>         ofst = i * 100
        >>>         inst_native_elements["R0_" + str(i)] = Rect(
        >>>             xy=[[ofst, 0], [ofst + 10, 10]], layer=["M1", "drawing"]
        >>>         )
        >>>         inst_native_elements["R1_" + str(i)] = Rect(
        >>>             xy=[[ofst + 90, 90], [ofst + 100, 100]], layer=["M1", "drawing"]
        >>>         )
        >>>     inst_native_elements["R2"] = Rect(
        >>>         xy=[[0, 0], [m * 100, 100]], layer=["prBoundary", "drawing"]
        >>>     )
        >>>     inst = VirtualInstance(
        >>>         name=name,
        >>>         libname="mylib",
        >>>         cellname="myvinst",
        >>>         xy=np.array([0, 0]),
        >>>         native_elements=inst_native_elements,
        >>>         shape=shape,
        >>>         pitch=pitch,
        >>>         unit_size=[m * 100, 100],
        >>>         pins=inst_pins,
        >>>         transform=transform,
        >>>         params=params,
        >>>     )
        >>>     return inst
        >>> # UserDefinedTemplate construction.
        >>> user_temp = UserDefinedTemplate(
        >>>     name="myusertemplate",
        >>>     bbox_func=user_bbox_func,
        >>>     pins_func=user_pins_func,
        >>>     generate_func=user_generate_func,
        >>> )
        >>> # VirtualInstance generation.
        >>> user_inst = user_temp.generate(name="myinst", params={"mult": 5})
        >>> # Display
        >>> print(user_temp)
            <laygo2.object.template.UserDefinedTemplate object at 0x00000192BF990130> 
            name: myusertemplate, class: UserDefinedTemplate,
        >>> print(user_inst)
            <laygo2.object.physical.VirtualInstance object at 0x00000192BF990280>
                name: myinst,
                class: VirtualInstance,
                xy: [0, 0],
                params: {'mult': 5},
                size: [500, 100],
                shape: [1, 1],
                pitch: [500, 100],
                transform: R0,
                pins: {'in0': <laygo2.object.physical.Pin object at 0x00000192BF9930D0>, 
                       'out0': <laygo2.object.physical.Pin object at 0x00000192BF9931C0>, 
                       'in1': <laygo2.object.physical.Pin object at 0x00000192BF993760>, 
                       'out1': <laygo2.object.physical.Pin object at 0x00000192BF9936A0>, 
                       'in2': <laygo2.object.physical.Pin object at 0x00000192BF993610>, 
                       'out2': <laygo2.object.physical.Pin object at 0x00000192BF9935B0>, 
                       'in3': <laygo2.object.physical.Pin object at 0x00000192BF9932E0>, 
                       'out3': <laygo2.object.physical.Pin object at 0x00000192BF9931F0>, 
                       'in4': <laygo2.object.physical.Pin object at 0x00000192BF993130>, 
                       'out4': <laygo2.object.physical.Pin object at 0x00000192BF9930A0>},
                native elements: {'R0_0': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_0': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_1': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_1': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_2': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_2': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_3': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_3': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R0_4': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R1_4': <laygo2.object.physical.Rect object at 0x0...>, 
                                  'R2': <laygo2.object.physical.Rect object at 0x000...>}
        >>> print(user_inst.bbox)
            [[  0   0]
             [500 100]]

        .. image:: ../assets/img/object_template_UserDefinedTemplate_generate.png
          :height: 250

        Notes
        -----
        **(Korean)** 템플릿으로부터 입력 파라미터에 따른 VirtualInstance 객체 생성.
        파라미터
            - name(str): 생성할 인스턴스의 이름.
            - shape(numpy.ndarray): (optional) 생성할 객체의 배열 shape.
            - pitch(numpy.ndarray): (optional) 생성할 객체 간의 간격.
            - params(dict) : 개체의 속성이 담긴 dictionary.
            - transform(str): 생성할 개체의 변환 속성.
        반환값
            - laygo2.VirtualInstance: 생성된 객체
        """
        inst = self._generate(
            name=name, shape=shape, pitch=pitch, transform=transform, params=params
        )
        # update netnames if netmap is provided.
        if netmap is not None:
            inst.update_netname(netmap=netmap)  
        return inst


# Test
if __name__ == "__main__":
    test_native_template = True
    test_pcell_template = True
    test_user_template = True

    import laygo2.object

    if test_native_template:
        print("NativeInstanceTemplate test")
        # define pins
        nat_temp_pins = dict()
        nat_temp_pins["in"] = laygo2.object.Pin(
            xy=[[0, 0], [10, 10]], layer=["M1", "drawing"], netname="in"
        )
        nat_temp_pins["out"] = laygo2.object.Pin(
            xy=[[90, 90], [100, 100]], layer=["M1", "drawing"], netname="out"
        )
        # create a template
        nat_temp = NativeInstanceTemplate(
            libname="mylib",
            cellname="mynattemplate",
            bbox=[[0, 0], [100, 100]],
            pins=nat_temp_pins,
        )
        # generate
        nat_inst = nat_temp.generate(
            name="mynatinst", shape=[2, 2], pitch=[100, 100], transform="R0"
        )
        # display
        print(nat_temp)
        print(nat_inst)

    if test_pcell_template:
        print("ParameterizedInstanceTemplate test")

        # define the bbox computation function.
        def pcell_bbox_func(params):
            return np.array([[0, 0], [100 * params["mult"], 100]])

        # define the pin generation function.
        def pcell_pins_func(params):
            template_pins = dict()
            for i in range(params["mult"]):
                template_pins["in" + str(i)] = laygo2.object.Pin(
                    xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
                    layer=["M1", "drawing"],
                    netname="in" + str(i),
                )
                template_pins["out" + str(i)] = laygo2.object.Pin(
                    xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
                    layer=["M1", "drawing"],
                    netname="out" + str(i),
                )
            return template_pins

        # create a template.
        pcell_temp = ParameterizedInstanceTemplate(
            libname="mylib",
            cellname="mypcelltemplate",
            bbox_func=pcell_bbox_func,
            pins_func=pcell_pins_func,
        )
        # generate based on the parameter assigned.
        pcell_inst_params = {"mult": 4}
        pcell_inst_size = pcell_temp.size(params=pcell_inst_params)
        pcell_inst = pcell_temp.generate(
            name="mypcellinst",
            shape=[2, 2],
            pitch=pcell_inst_size,
            transform="R0",
            params=pcell_inst_params,
        )
        # display
        print(pcell_temp)
        print(pcell_inst)

    if test_user_template:
        print("UserDefinedTemplate test")

        # define the bbox computation function.
        def user_bbox_func(params):
            return np.array([[0, 0], [100 * params["mult"], 100]])

        # define the pin generation function.
        def user_pins_func(params):
            template_pins = dict()
            for i in range(params["mult"]):
                template_pins["in" + str(i)] = laygo2.object.Pin(
                    xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
                    layer=["M1", "drawing"],
                    netname="in" + str(i),
                )
                template_pins["out" + str(i)] = laygo2.object.Pin(
                    xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
                    layer=["M1", "drawing"],
                    netname="out" + str(i),
                )
            return template_pins

        # define the instance generation function.
        def user_generate_func(
            name=None, shape=None, pitch=np.array([0, 0]), transform="R0", params=None
        ):
            m = params["mult"]
            shape = np.array([1, 1]) if shape is None else np.asarray(shape)

            inst_pins = user_pins_func(params)
            inst_native_elements = dict()
            for i in range(m):
                ofst = i * 100
                inst_native_elements["R0_" + str(i)] = laygo2.object.Rect(
                    xy=[[ofst, 0], [ofst + 10, 10]], layer=["M1", "drawing"]
                )
                inst_native_elements["R1_" + str(i)] = laygo2.object.Rect(
                    xy=[[ofst + 90, 90], [ofst + 100, 100]], layer=["M1", "drawing"]
                )
            inst_native_elements["R2"] = laygo2.object.Rect(
                xy=[[0, 0], [m * 100, 100]], layer=["prBoundary", "drawing"]
            )
            inst = laygo2.object.VirtualInstance(
                name=name,
                libname="mylib",
                cellname="myvinst",
                xy=np.array([0, 0]),
                native_elements=inst_native_elements,
                shape=shape,
                pitch=pitch,
                unit_size=[m * 100, 100],
                pins=inst_pins,
                transform=transform,
                params=params,
            )
            return inst

        user_temp = UserDefinedTemplate(
            name="myusertemplate",
            bbox_func=user_bbox_func,
            pins_func=user_pins_func,
            generate_func=user_generate_func,
        )
        user_inst = user_temp.generate(
            name="myuserinst", shape=[2, 1], params={"mult": 5}
        )
        print(user_temp)
        print(user_inst)
        print(user_inst.bbox)



# -*- coding: utf-8 -*-
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""Grid library for the advanced example technology (advtech)."""

import laygo2.object.database
import laygo2.object.grid
#Should be `import laygo2_tech as tech` for actual use.
#import laygo2.examples.laygo2_tech as tech
import laygo2_tech_quick_start as tech
#import laygo2_tech as tech


tech_params = tech.tech_params

# Grid library
def load_grids(templates):
    """
    Load grids to a grid library object.

    Parameters
    ----------
    templates: laygo2.object.database.TemplateLibrary
        The template library object that contains via templates.
    """
    glib = laygo2.object.database.GridLibrary(name='advtech_grids')
    for gname, grid in tech_params['grids'].items():
        gv = laygo2.object.grid.OneDimGrid(name=gname + '_v', scope=grid['vertical']['scope'],
                                           elements=grid['vertical']['elements'])
        gh = laygo2.object.grid.OneDimGrid(name=gname + '_h', scope=grid['horizontal']['scope'],
                                           elements=grid['horizontal']['elements'])
        if grid['type'] == 'placement':  # placement grid
            g = laygo2.object.grid.PlacementGrid(name=gname, vgrid=gv, hgrid=gh)
            glib.append(g)
        elif grid['type'] == 'routing':  # routing grid
            vwidth = laygo2.object.grid.CircularMapping(elements=grid['vertical']['width'])
            hwidth = laygo2.object.grid.CircularMapping(elements=grid['horizontal']['width'])
            vextension = laygo2.object.grid.CircularMapping(elements=grid['vertical']['extension'])
            hextension = laygo2.object.grid.CircularMapping(elements=grid['horizontal']['extension'])
            vlayer = laygo2.object.grid.CircularMapping(elements=grid['vertical']['layer'], dtype=object)
            hlayer = laygo2.object.grid.CircularMapping(elements=grid['horizontal']['layer'], dtype=object)
            pin_vlayer = laygo2.object.grid.CircularMapping(elements=grid['vertical']['pin_layer'], dtype=object)
            pin_hlayer = laygo2.object.grid.CircularMapping(elements=grid['horizontal']['pin_layer'], dtype=object)
            primary_grid = grid['primary_grid']
            # Create the via map defined by the yaml file.
            vmap_original = grid['via']['map']  # viamap defined in the yaml file.
            vmap_mapped = list()  # map template objects to the via map.
            for vmap_org_row in vmap_original:
                vmap_mapped_row = []
                for vmap_org_elem in vmap_org_row:
                    vmap_mapped_row.append(templates[vmap_org_elem])
                vmap_mapped.append(vmap_mapped_row)
            viamap = laygo2.object.grid.CircularMappingArray(elements=vmap_mapped, dtype=object)

            g = laygo2.object.grid.RoutingGrid(name=gname, vgrid=gv, hgrid=gh,
                                               vwidth=vwidth, hwidth=hwidth,
                                               vextension=vextension, hextension=hextension,
                                               vlayer=vlayer, hlayer=hlayer,
                                               pin_vlayer=pin_vlayer, pin_hlayer=pin_hlayer,
                                               viamap=viamap, primary_grid=primary_grid)
            glib.append(g)
    return glib




#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements the interface with matplotlib.

"""
import logging
from math import log10
from decimal import *

import numpy as np
import laygo2.object
import laygo2.util.transform as tf

import matplotlib
import matplotlib.pyplot as plt

__author__ = ""
__maintainer__ = ""
__status__ = "Prototype"


def _translate_obj(objname, obj, colormap, scale=1, master=None, offset=np.array([0, 0])):
    """
    Convert a layout object to corresponding matplotlib patch object.
    """
    if master is None:
        mxy = np.array([0, 0])
        mtf = "R0"
    else:  # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform
    if obj.__class__ == laygo2.object.Rect:
        # Draw a rectangle
        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(
            _xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
            tf.Mt(mtf).T,
        )
        size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
        if obj.layer[0] in colormap:
            rect = matplotlib.patches.Rectangle(
                (_xy[0, 0], _xy[0, 1]),
                size[0],
                size[1],
                facecolor=colormap[obj.layer[0]][1],
                edgecolor=colormap[obj.layer[0]][0],
                alpha=colormap[obj.layer[0]][2],
                lw=2,
            )
            # ax.add_patch(rect)
            return [[rect, obj.layer[0]]]
        return []
    elif obj.__class__ == laygo2.object.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        for idx, _obj in np.ndenumerate(_objelem):
            # Invoke _laygo2_generate_pin(cv, name, layer, bbox) in {header_filename}
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            size = [_xy[1, 0] - _xy[0, 0], _xy[1, 1] - _xy[0, 1]]
            if obj.layer[0] in colormap:
                rect = matplotlib.patches.Rectangle(
                    (_xy[0, 0], _xy[0, 1]),
                    size[0],
                    size[1],
                    facecolor=colormap[obj.layer[0]][1],
                    edgecolor=colormap[obj.layer[0]][0],
                    alpha=colormap[obj.layer[0]][2],
                    lw=2,
                )
                return [[rect, obj.layer[0], objname+"/"+obj.netname]]
                # ax.add_patch(rect)
            return []
    elif obj.__class__ == laygo2.object.Text:
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.Instance:
        # Invoke _laygo2_generate_instance( cv name libname cellname viewname loc orient num_rows num_cols
        # sp_rows sp_cols params params_order )
        _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
        if master is None:
            transform = obj.transform
        else:  # if the translated object has a master (e.g. VirtualInstance)
            transform = tf.combine(obj.transform, master.transform)
        if obj.shape is None:
            num_rows = 1
            num_cols = 1
            sp_rows = 0
            sp_cols = 0
        else:
            num_rows = obj.shape[1]
            num_cols = obj.shape[0]
            sp_rows = obj.pitch[1]
            sp_cols = obj.pitch[0]

        _xy0 = obj.xy0
        _xy1 = np.dot(obj.size, tf.Mt(mtf).T) * np.array([num_rows, num_cols])
        rect = matplotlib.patches.Rectangle(
            (_xy0[0], _xy0[1]), _xy1[0], _xy1[1], facecolor=colormap["__instance__"][1],
            edgecolor=colormap["__instance__"][0], alpha=colormap["__instance__"][2], lw=2
        )
        pypobjs = [[rect, "__instance__", obj.cellname+"/"+obj.name]]

        # Instance pins
        for pn, p in obj.pins.items():
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=master, offset=offset)
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            pypobjs += _pypobj
        return pypobjs
    elif obj.__class__ == laygo2.object.VirtualInstance:
        pypobjs = []
        if obj.shape is None:
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.Pin:
                    if obj.name == None:
                        objname = "NoName"
                    else:
                        objname = obj.name
                    _pypobj = _translate_obj(objname + "_" + elem_name, elem, colormap, scale=scale, master=obj)
                    pypobjs += _pypobj
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.Pin:
                        _pypobj = _translate_obj(
                            obj.name + "_" + elem_name + str(i) + "_" + str(j),
                            elem,
                            colormap,
                            scale=scale,
                            master=obj[i, j],
                        )
                        pypobjs += _pypobj
        
        # Instance pins
        for pn, p in obj.pins.items():
            _pypobj = _translate_obj(pn, p, colormap, scale=scale, master=master, offset=offset)
            _pypobj[0][1] = "__instance_pin__"
            _pypobj[0][0].set(edgecolor=colormap["__instance_pin__"][0])
            _pypobj[0][0].set(facecolor=colormap["__instance_pin__"][1])
            pypobjs += _pypobj
        return pypobjs

    else:
        return []
        #return [obj.translate_to_matplotlib()]
    return []


def export(
    db,
    cellname=None,
    scale=1,
    colormap=None,
    order=None,
    xlim=[-100, 400],
    ylim=[-100, 300],
):
    """
    Export a laygo2.object.database.Library object to a matplotlib plot.

    Parameters
    ----------
    db: laygo2.database.Library
        The library database to exported.
    cellname: str or List[str]
        The name(s) of cell(s) to be exported.
    scale: float
        The scaling factor between laygo2's integer coordinates and plot coordinates.

    Returns
    -------
    matplotlib.pyplot.figure or list: The generated figure object(s).

    """
    # colormap
    if colormap is None:
        colormap = dict()

    # a list to align layered objects in order
    if order is None:
        order = []

    # cell name handling.
    cellname = db.keys() if cellname is None else cellname  # export all cells if cellname is not given.
    cellname = [cellname] if isinstance(cellname, str) else cellname  # convert to a list for iteration.

    fig = []
    for cn in cellname:
        _fig = plt.figure()
        pypobjs = []
        ax = _fig.add_subplot(111)
        for objname, obj in db[cn].items():
            pypobjs += _translate_obj(objname, obj, colormap, scale=scale)
        for _alignobj in order:
            for _pypobj in pypobjs:
                if _pypobj[1] == _alignobj:  # layer is matched.
                    if _pypobj[0].__class__ == matplotlib.patches.Rectangle:  # Rect
                        ax.add_patch(_pypobj[0])
                        if len(_pypobj) == 3:  # annotation.
                            ax.add_artist(_pypobj[0])
                            rx, ry = _pypobj[0].get_xy()
                            cx = rx + _pypobj[0].get_width()/2.0
                            cy = ry + _pypobj[0].get_height()/2.0
                            if _pypobj[1] == "__instance_pin__":
                                color = _pypobj[0].get_edgecolor()
                            else:
                                color = "black"
                            ax.annotate(_pypobj[2], (cx, cy), color=color, weight='bold', fontsize=6, ha='center', va='center')
        fig.append(_fig)
    if len(fig) == 1:
        fig = fig[0]

    plt.xlim(xlim)
    plt.ylim(ylim)
    plt.show()

    return fig



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2020, Nifty Chips Laboratory, Hanyang University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interfaces with yaml files.

"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

import yaml
import os.path
import laygo2

def export_template(template, filename, mode='append'):
    """Export a template to a yaml file.

    Parameters
    ----------
    template: laygo2.object.template.Template
        The template object to be exported.        
    filename: str
        The name of the yaml file.
    mode: str
        If 'append', it adds the template entry without erasing the 
        preexisting file.

    Example
    -------
    >>> import laygo2
    >>> from laygo2.object.physical import Pin
    >>> from laygo2.object.template import NativeInstanceTemplate
    >>> p = dict()
    >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'],
    >>>                  netname='i')
    >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'],
    >>>                  netname='o')
    >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp',
    >>>                                 bbox=[[0, 0], [100, 100]], pins=p)
    >>> laygo2.interface.yaml.export_template(nt, filename="mytemplates.yaml")    
    Your design was translated into YAML format.
    {'mylib': {
        'mytemp': {
            'libname': 'mylib', 
            'cellname': 'mytemp', 
            'bbox': [[0, 0], [100, 100]], 
            'pins': {
                'i': {
                    'xy': [[0, 0], [10, 10]], 
                    'layer': ['M1', 'drawing'], 
                    'name': None, 
                    'netname': 'i'
                    }, 
                'o': {
                    'xy': [[90, 90], [100, 100]], 
                    'layer': ['M1', 'drawing'], 
                    'name': None, 
                    'netname': 'o'
    }}}}}
    """
    libname = template.libname
    cellname = template.cellname
    pins = template.pins()

    db = dict()
    if mode == 'append':  # in append mode, the template is appended to 'filename' if the file exists.
        if os.path.exists(filename):
            with open(filename, 'r') as stream:
                db = yaml.load(stream, Loader=yaml.FullLoader)
    if libname not in db:
        db[libname] = dict()
    db[libname][cellname] = template.export_to_dict()
    with open(filename, 'w') as stream:
        yaml.dump(db, stream)
    #print("Your design was translated into YAML format.")
    return db

#filename=libname+'_templates.yaml'

def import_template(filename):
    """Import templates from a yaml file.

    Parameters
    ----------
    filename: str
        The name of the yaml file.
    
    Example
    -------
    >>> import laygo2
    >>> from laygo2.object.physical import Pin
    >>> from laygo2.object.template import NativeInstanceTemplate
    >>> p = dict()
    >>> p['i'] = Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'],
    >>>                  netname='i')
    >>> p['o'] = Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'],
    >>>                  netname='o')
    >>> nt = NativeInstanceTemplate(libname='mylib', cellname='mytemp',
    >>>                                 bbox=[[0, 0], [100, 100]], pins=p)
    >>> laygo2.interface.yaml.export_template(nt, filename="mytemplates.yaml")    
    >>> # Import the template back to python.
    >>> my_tlib = laygo2.interface.yaml.import_template("mytemplates.yaml")
    >>> print(my_tlib)
    <laygo2.object.database.TemplateLibrary object at 0x000001FE3440A410> 
        name: mylib, 
        params: None       
        elements: {
            'mytemp': <laygo2.object.template.NativeInstanceTemplate object at 0x000001FE3440A2C0>
        }
    """
    # load yaml file
    if os.path.exists(filename):
        with open(filename, 'r') as stream:
            db = yaml.load(stream, Loader=yaml.FullLoader)
    libname = list(db.keys())[0]  # assuming there's only one library defined in each file.
    # create template library
    tlib = laygo2.object.database.TemplateLibrary(name=libname)
    # read out the yaml file entries and build template objects
    for tn, tdict in db[libname].items():
        pins = dict()
        if 'pins' in tdict:
            for pinname, pdict in tdict['pins'].items():
                pins[pinname] = laygo2.object.Pin(xy=pdict['xy'], layer=pdict['layer'], netname=pdict['netname'])
        t = laygo2.object.NativeInstanceTemplate(libname=libname, cellname=tn, bbox=tdict['bbox'], pins=pins)
        tlib.append(t)
    return tlib

